#define _GNU_SOURCE

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include "SDL.h"
#include "SDL_image.h"

#include "fieldsat.h"

#define SCREEN_WIDTH 400
#define SCREEN_HEIGHT 400

/*
#define DRAW_ARROWS
*/

static void init(void);

static void update(void);

static void display(void);

static void draw_grid(void);

static void model_sat(void);

static GLdouble length(GLdouble x, GLdouble y, GLdouble z);

static void draw_arrow(GLdouble x, GLdouble y, GLdouble z);

static void draw_length_arrow(GLdouble x, GLdouble y, GLdouble z, GLdouble len);

static void handle_key(SDL_KeyboardEvent *key);

/* PowderBlue from X11 colors. */
static const GLubyte CLEAR_COLOR[3] = { 176, 224, 230 };
/*
static const GLubyte CLEAR_COLOR[3] = { 20, 20, 20 };
*/
/* Beige from X11 colors. */
static const GLubyte SAT_COLOR[3] = { 245, 245, 200 };
static const GLubyte SAT_LINE_COLOR[3] = { 245, 0, 245 };

/*
static const GLfloat GLOBAL_AMBIENT[4] = { 0.8, 0.8, 0.8, 0.8 };
*/
static const GLfloat GLOBAL_AMBIENT[4] = { 0.2, 0.2, 0.2, 0.2 };
static const GLfloat AMBIENT[4] = { 0.4, 0.4, 0.4, 0.4 };
static const GLfloat DIFFUSE[4] = { 0.4, 0.4, 0.4, 0.4 };
static const GLfloat SPECULAR[4] = { 0.2, 0.2, 0.2, 0.2 };
static const GLfloat POSITION[4] = { -1.0, 0.2, 0.3, 0.0 };

static FILE *fp;

static struct fieldsat_state state;
static double current_time, current_sample_time;
static unsigned long previous_timestamp;

static unsigned int quit = 0;

static double time_multiplier = 1.0;

static GLuint texture_names[2];

static void init(void)
{
	SDL_Surface *texture;

	glClearColor(CLEAR_COLOR[0] / 255.0,
		CLEAR_COLOR[1] / 255.0,
		CLEAR_COLOR[2]/ 255.0,
		0.0);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glFrustum(-0.1, 0.1, -0.1, 0.1, 1.0, 1000.0);

	glEnable(GL_DEPTH_TEST);

	glEnable(GL_LIGHTING);
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, GLOBAL_AMBIENT);
	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);

	glLightfv(GL_LIGHT0, GL_AMBIENT, AMBIENT);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, DIFFUSE);
	glLightfv(GL_LIGHT0, GL_SPECULAR, SPECULAR);
	glLightfv(GL_LIGHT0, GL_POSITION, POSITION);
	glEnable(GL_LIGHT0);

	glEnable(GL_COLOR_MATERIAL);

	glEnable(GL_NORMALIZE);

	glEnable(GL_TEXTURE_2D);
	glGenTextures(2, texture_names);

	texture = IMG_Load("textures-1.jpg");
	if (texture == NULL)
		fprintf(stderr, "Error loading texture.\n");
	glBindTexture(GL_TEXTURE_2D, texture_names[0]);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexImage2D(GL_TEXTURE_2D, 0, 3, texture->w, texture->h, 0, GL_RGB, GL_UNSIGNED_BYTE, texture->pixels);
	SDL_FreeSurface(texture);
	texture = IMG_Load("textures-2.jpg");
	if (texture == NULL)
		fprintf(stderr, "Error loading texture.\n");
	glBindTexture(GL_TEXTURE_2D, texture_names[1]);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexImage2D(GL_TEXTURE_2D, 0, 3, texture->w, texture->h, 0, GL_RGB, GL_UNSIGNED_BYTE, texture->pixels);
	SDL_FreeSurface(texture);

	state_init(&state);

	synchronize(fp);
}

static void update(void)
{
	struct fieldsat_sample sample;
	double t;

	while (current_sample_time < current_time) {
		if (!read_sample(fp, &sample)) {
			synchronize(fp);
			if (!read_sample(fp, &sample)) {
				fprintf(stderr, "Sample read error.\n");
				exit(1);
			}
		}

		/*
		sample_print(&sample);
		*/

		if (sample.type == SAMPLE_TYPE_GO) {
			state_init(&state);
			current_sample_time = 0.0;
			current_time = 0.0;
		}

		if (current_sample_time == 0.0)
			t = current_time;
		else
			t = (double) (sample.timestamp - previous_timestamp) / F_CPU;

		if (t > 1.0)
			fprintf(stderr, "Warning, time delta = %f.\n", t);

		current_sample_time += t;

		previous_timestamp = sample.timestamp;

		state_update(&state, &sample, t);
	}
}

static void display(void)
{
	static double rotation_mat[16] = {
		1, 0, 0, 0,
		0, 1, 0, 0,
		0, 0, 1, 0,
		0, 0, 0, 1
	};
	static unsigned char buf[SCREEN_WIDTH * SCREEN_HEIGHT * 3];
	GLfloat ambient[4], diffuse[4], specular[4];
	unsigned int i, j;

	glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);

	for (i = 0; i < 4; i++) {
		ambient[i] = state.solar_luminosity * AMBIENT[i];
		diffuse[i] = state.solar_luminosity * DIFFUSE[i];
		specular[i] = state.solar_luminosity * SPECULAR[i];
	}
	glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);
	glLightfv(GL_LIGHT0, GL_SPECULAR, specular);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	/*
	glTranslated(0.0, 0.0, -3.0);
	gluLookAt(0.0, 0.0, 3.0, state.position[0], state.position[1], state.position[2], 0.0, 1.0, 0.0);
	gluLookAt(0.0, 0.0, 3.0, 0.0, state.position[1], 0.0, 0.0, 1.0, 0.0);
	*/
	gluLookAt(0.0, 0.0, 3.0, state.position[0], 0.0, state.position[2], 0.0, 1.0, 0.0);

	/*
	draw_grid();
	*/

	glPushMatrix();
	/*
	glTranslated(state.position[0], state.position[1], state.position[2]);
	glTranslated(0.0, state.position[1], 0.0);
	*/
	glTranslated(state.position[0], 0.0, state.position[2]);
	/*
	glRotated(state.rotation[0], 1.0, 0.0, 0.0);
	glRotated(state.rotation[1], 0.0, 1.0, 0.0);
	glRotated(state.rotation[2], 0.0, 0.0, 1.0);
	*/
	/*
	{
	double len = length(state.acceleration[0], state.acceleration[1], state.acceleration[2]);
	glRotated(acos(-state.acceleration[0] / len) * 180.0 / 3.1415926, 0.0, state.acceleration[2], -state.acceleration[1]);
	}
	*/

	glPushMatrix();

	{
	double s = state.rotation[0], x = state.rotation[1],
		y = state.rotation[2], z = state.rotation[3];

	rotation_mat[0] = 1 - 2 * y * y - 2 * z * z;
	rotation_mat[1] = 2 * x * y + 2 * s * z;
	rotation_mat[2] = 2 * x * z - 2 * s * y;
	rotation_mat[4] = 2 * x * y - 2 * s * z;
	rotation_mat[5] = 1 - 2 * x * x - 2 * z * z;
	rotation_mat[6] = 2 * y * z + 2 * s * x;
	rotation_mat[8] = 2 * x * z + 2 * s * y;
	rotation_mat[9] = 2 * y * z - 2 * s * x;
	rotation_mat[10] = 1 - 2 * x * x - 2 * y * y;
	glMultMatrixd(rotation_mat);
	}

	glPushMatrix();

	/* Translate the satellite by the scaled difference between the current
	   and average acceleration vectors. */
	glTranslated(-(state.acceleration[0] - state.avg_acc[0]) / 250.0, -(state.acceleration[1] - state.avg_acc[1]) / 250.0, -(state.acceleration[2] - state.avg_acc[2]) / 250.0);

	model_sat();
	/*
	*/

	glPopMatrix();

#ifdef DRAW_ARROWS
	glClear(GL_DEPTH_BUFFER_BIT);

	glColor3d(0.0, 1.0, 0.0);
	draw_length_arrow(0.0, -1.0, 0.0, 0.2);
#endif

	glPopMatrix();

#ifdef DRAW_ARROWS
	glColor3d(1.0, 0.0, 0.0);
	draw_arrow(-state.acceleration[0] / 9.8 * 0.2, -state.acceleration[1] / 9.8 * 0.2, -state.acceleration[2] / 9.8 * 0.2);

	/*
	glColor3d(1.0, 0.0, 1.0);
	draw_arrow(state.velocity[0] / 9.8, state.velocity[1] / 9.8, state.velocity[2] / 9.8);

	glColor3d(0.0, 0.0, 1.0);
	draw_arrow(state.lin_acc[0] / 9.8 * 0.2, state.lin_acc[1] / 9.8 * 0.2, state.lin_acc[2] / 9.8 * 0.2);
	*/
#endif
	glPopMatrix();

	SDL_GL_SwapBuffers();

	glReadPixels(0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, GL_RGB, GL_UNSIGNED_BYTE, buf);
	for (j = 0; j < SCREEN_HEIGHT; j++) {
		for (i = 0; i < SCREEN_WIDTH * 3; i++)
			putchar(buf[(SCREEN_HEIGHT - j - 1) * SCREEN_WIDTH * 3 + i]);
	}
}

static void draw_grid(void)
{
	int i;

	glBegin(GL_LINES);
	for (i = -5; i <= 5; i++) {
		glVertex3d(i, -5, 0);
		glVertex3d(i, 5, 0);
		glVertex3d(-5, i, 0);
		glVertex3d(5, i, 0);
	}
	glEnd();
}

static void model_sat(void)
{
	const double HEIGHT = 0.26, WIDTH = 0.24, DEPTH = 0.22;
	GLint mode[2];

	glGetIntegerv(GL_POLYGON_MODE, mode);

	if (mode[0] == GL_FILL) {
		if (glIsEnabled(GL_TEXTURE_2D))
			glColor3d(1.0, 1.0, 1.0);
		else
			glColor3ubv(SAT_COLOR);
	} else {
		glColor3ubv(SAT_LINE_COLOR);
	}

	glBindTexture(GL_TEXTURE_2D, texture_names[1]);

	glBegin(GL_TRIANGLE_STRIP);
	glNormal3d(-1.0, 0.0, 0.0);
	glTexCoord2d(0.0, 0.5);
	glVertex3d(-WIDTH / 2.0, -HEIGHT / 2.0, -DEPTH / 2.0);
	glTexCoord2d(0.5, 0.5);
	glVertex3d(-WIDTH / 2.0, -HEIGHT / 2.0, DEPTH / 2.0);
	glTexCoord2d(0.0, 0.0);
	glVertex3d(-WIDTH / 2.0, HEIGHT / 2.0, -DEPTH / 2.0);
	glTexCoord2d(0.5, 0.0);
	glVertex3d(-WIDTH / 2.0, HEIGHT / 2.0, DEPTH / 2.0);
	glEnd();
	glBegin(GL_TRIANGLE_STRIP);
	glNormal3d(1.0, 0.0, 0.0);
	glTexCoord2d(0.5, 0.5);
	glVertex3d(WIDTH / 2.0, -HEIGHT / 2.0, DEPTH / 2.0);
	glTexCoord2d(1.0, 0.5);
	glVertex3d(WIDTH / 2.0, -HEIGHT / 2.0, -DEPTH / 2.0);
	glTexCoord2d(0.5, 0.0);
	glVertex3d(WIDTH / 2.0, HEIGHT / 2.0, DEPTH / 2.0);
	glTexCoord2d(1.0, 0.0);
	glVertex3d(WIDTH / 2.0, HEIGHT / 2.0, -DEPTH / 2.0);
	glEnd();

	glPushMatrix();
	glTranslated(-WIDTH / 2.0, HEIGHT / 2.0, 0.0);
	glRotated(state.solar_is_deployed ? -85.0 : -10.0, 0.0, 0.0, 1.0);
	glBegin(GL_TRIANGLE_STRIP);
	glTexCoord2d(0.5, 0.5);
	glVertex3d(0.0, 0.0, -DEPTH * 0.8 / 2.0);
	glTexCoord2d(0.0, 0.5);
	glVertex3d(0.0, -HEIGHT, -DEPTH * 0.8 / 2.0);
	glTexCoord2d(0.5, 1.0);
	glVertex3d(0.0, 0.0, DEPTH * 0.8 / 2.0);
	glTexCoord2d(0.0, 1.0);
	glVertex3d(0.0, -HEIGHT, DEPTH * 0.8 / 2.0);
	glEnd();
	glPopMatrix();
	glPushMatrix();
	glTranslated(WIDTH / 2.0, HEIGHT / 2.0, 0.0);
	glRotated(state.solar_is_deployed ? 85.0 : 10.0, 0.0, 0.0, 1.0);
	glBegin(GL_TRIANGLE_STRIP);
	glTexCoord2d(0.5, 0.5);
	glVertex3d(0.0, 0.0, DEPTH * 0.8 / 2.0);
	glTexCoord2d(0.0, 0.5);
	glVertex3d(0.0, -HEIGHT, DEPTH * 0.8 / 2.0);
	glTexCoord2d(0.5, 1.0);
	glVertex3d(0.0, 0.0, -DEPTH * 0.8 / 2.0);
	glTexCoord2d(0.0, 1.0);
	glVertex3d(0.0, -HEIGHT, -DEPTH * 0.8 / 2.0);
	glEnd();
	glPopMatrix();

	glBindTexture(GL_TEXTURE_2D, texture_names[0]);

	glBegin(GL_TRIANGLE_STRIP);
	glNormal3d(0.0, -1.0, 0.0);
	glTexCoord2d(0.5, 0.5);
	glVertex3d(-WIDTH / 2.0, -HEIGHT / 2.0, -DEPTH / 2.0);
	glTexCoord2d(1.0, 0.5);
	glVertex3d(WIDTH / 2.0, -HEIGHT / 2.0, -DEPTH / 2.0);
	glTexCoord2d(0.5, 0.0);
	glVertex3d(-WIDTH / 2.0, -HEIGHT / 2.0, DEPTH / 2.0);
	glTexCoord2d(1.0, 0.0);
	glVertex3d(WIDTH / 2.0, -HEIGHT / 2.0, DEPTH / 2.0);
	glEnd();
	glBegin(GL_TRIANGLE_STRIP);
	glNormal3d(0.0, 1.0, 0.0);
	glTexCoord2d(0.0, 0.5);
	glVertex3d(-WIDTH / 2.0, HEIGHT / 2.0, DEPTH / 2.0);
	glTexCoord2d(0.5, 0.5);
	glVertex3d(WIDTH / 2.0, HEIGHT / 2.0, DEPTH / 2.0);
	glTexCoord2d(0.0, 0.0);
	glVertex3d(-WIDTH / 2.0, HEIGHT / 2.0, -DEPTH / 2.0);
	glTexCoord2d(0.5, 0.0);
	glVertex3d(WIDTH / 2.0, HEIGHT / 2.0, -DEPTH / 2.0);
	glEnd();

	glBegin(GL_TRIANGLE_STRIP);
	glNormal3d(0.0, 0.0, -1.0);
	glTexCoord2d(1.0, 0.5);
	glVertex3d(-WIDTH / 2.0, HEIGHT / 2.0, -DEPTH / 2.0);
	glTexCoord2d(0.5, 0.5);
	glVertex3d(WIDTH / 2.0, HEIGHT / 2.0, -DEPTH / 2.0);
	glTexCoord2d(1.0, 1.0);
	glVertex3d(-WIDTH / 2.0, -HEIGHT / 2.0, -DEPTH / 2.0);
	glTexCoord2d(0.5, 1.0);
	glVertex3d(WIDTH / 2.0, -HEIGHT / 2.0, -DEPTH / 2.0);
	glEnd();
	glBegin(GL_TRIANGLE_STRIP);
	glNormal3d(0.0, 0.0, 1.0);
	glTexCoord2d(0.0, 1.0);
	glVertex3d(-WIDTH / 2.0, -HEIGHT / 2.0, DEPTH / 2.0);
	glTexCoord2d(0.5, 1.0);
	glVertex3d(WIDTH / 2.0, -HEIGHT / 2.0, DEPTH / 2.0);
	glTexCoord2d(0.0, 0.5);
	glVertex3d(-WIDTH / 2.0, HEIGHT / 2.0, DEPTH / 2.0);
	glTexCoord2d(0.5, 0.5);
	glVertex3d(WIDTH / 2.0, HEIGHT / 2.0, DEPTH / 2.0);
	glEnd();
}

static GLdouble length(GLdouble x, GLdouble y, GLdouble z)
{
	return sqrt(x * x + y * y + z * z);
}

static void draw_arrow(GLdouble x, GLdouble y, GLdouble z)
{
	GLdouble len;

	glPushAttrib(GL_ENABLE_BIT);
	glDisable(GL_TEXTURE_2D);

	len = length(x, y, z);
	glPushMatrix();
	glRotated(acos(x / len) * 180.0 / 3.1415926, 0.0, -z, y);
	glBegin(GL_LINES);
	glVertex3d(0.0, 0.0, 0.0);
	glVertex3d(len, 0.0, 0.0);
	glVertex3d(len - 0.01, -0.01, 0.0);
	glVertex3d(len, 0.0, 0.0);
	glVertex3d(len - 0.01, 0.01, 0.0);
	glVertex3d(len, 0.0, 0.0);
	glVertex3d(len - 0.01, 0.0, -0.01);
	glVertex3d(len, 0.0, 0.0);
	glVertex3d(len - 0.01, 0.0, 0.01);
	glVertex3d(len, 0.0, 0.0);
	glEnd();
	glBegin(GL_TRIANGLE_FAN);
	glNormal3d(1.0, 0.0, 0.0);
	glVertex3d(len, 0.0, 0.0);
	glNormal3d(0.0, 1.0, 0.0);
	glVertex3d(0.0, 0.005, 0.0);
	glNormal3d(0.0, 0.0, 1.0);
	glVertex3d(0.0, 0.0, 0.005);
	glNormal3d(0.0, -1.0, 0.0);
	glVertex3d(0.0, -0.005, 0.0);
	glNormal3d(0.0, 0.0, -1.0);
	glVertex3d(0.0, 0.0, -0.005);
	glNormal3d(0.0, 1.0, 0.0);
	glVertex3d(0.0, 0.005, 0.0);
	glEnd();
	glPopMatrix();

	glPopAttrib();
}

static void draw_length_arrow(GLdouble x, GLdouble y, GLdouble z, GLdouble len)
{
	len = length(x, y, z) / len;
	draw_arrow(x / len, y / len, z / len);
}

#ifdef main
#undef main
#endif

int main(int argc, char *argv[])
{
	SDL_Surface *screen;
	SDL_Event event;
	Uint32 ticks, ticks_delta;

	fp = stream_open(argc, argv);

	if (SDL_Init(SDL_INIT_VIDEO) == -1) {
		fprintf(stderr, "Can't initialize SDL: %s\n", SDL_GetError());
		return 1;
	}
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	screen = SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT, 32,
		SDL_ANYFORMAT | SDL_OPENGL);
	if (screen == NULL) {
		fprintf(stderr, "Can't set video mode: %s\n", SDL_GetError());
		return 1;
	}

	init();

	ticks = SDL_GetTicks();
	while (!quit) {
		ticks_delta = SDL_GetTicks() - ticks;
		ticks += ticks_delta;
		current_time += 1.0 / 30.0;
		update();
		display();
		while (SDL_PollEvent(&event)) {
			if (event.type == SDL_KEYDOWN
				|| event.type == SDL_KEYUP)
				handle_key(&event.key);
			else if (event.type == SDL_QUIT)
				quit = 1;
		}
	}

	SDL_FreeSurface(screen);
	SDL_Quit();

	return 0;
}

static void handle_key(SDL_KeyboardEvent *key)
{
	if (key->state == SDL_PRESSED) {
		if (key->keysym.sym == SDLK_EQUALS) {
			if (time_multiplier == -1.0)
				time_multiplier = 0.0;
			else if (time_multiplier == 0.0)
				time_multiplier = 1.0;
			else if (time_multiplier < 0.0)
				time_multiplier /= 2.0;
			else
				time_multiplier *= 2.0;
		} else if (key->keysym.sym == SDLK_MINUS) {
			if (time_multiplier == 1.0)
				time_multiplier = 0.0;
			else if (time_multiplier == 0.0)
				time_multiplier = -1.0;
			else if (time_multiplier > 0.0)
				time_multiplier /= 2.0;
			else
				time_multiplier *= 2.0;
		} else if (key->keysym.sym == SDLK_SPACE) {
			state_print(&state);
		} else if (key->keysym.sym == SDLK_ESCAPE) {
			quit = 1;
		}
	}
}
